package org.jastadd.relast.compiler;

import beaver.Parser;
import org.jastadd.option.ValueOption;
import org.jastadd.relast.ast.GrammarFile;
import org.jastadd.relast.ast.Program;
import org.jastadd.relast.parser.RelAstParser;
import org.jastadd.relast.scanner.RelAstScanner;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;

public class RelastSourceToSourceCompiler extends AbstractCompiler {


  protected ValueOption optionOutputDir;
  protected ValueOption optionInputDir;

  public RelastSourceToSourceCompiler(String name, boolean jastAddCompliant) {
    super(name, jastAddCompliant);
  }

  public static void main(String[] args) {
    try {
      new RelastSourceToSourceCompiler("relast-preprocessor", true).run(args);
    } catch (CompilerException e) {
      System.err.println(e.getMessage());
      System.exit(-1);
    }
  }

  @Override
  protected void initOptions() {
    optionOutputDir = addOption(new ValueOption("outputDir", "target directory for the generated files."));
    optionInputDir = addOption(new ValueOption("inputDir", "input directory."));
  }

  @Override
  protected int compile() throws CompilerException {
    Path inputPath;
    if (optionInputDir.isMatched()) {
      inputPath = Paths.get(optionInputDir.value());
    } else {
      inputPath = Paths.get(".");
      printMessage("No input dir is set. Assuming current directory '" + inputPath.toAbsolutePath().toString() + "'.");
    }

    if (!inputPath.toFile().exists()) {
      printMessage("Input path '" + inputPath.toAbsolutePath().toString() + "' does not exist. Exiting...");
      System.exit(-1);
    } else if (!inputPath.toFile().isDirectory()) {
      printMessage("Input path '" + inputPath.toAbsolutePath().toString() + "' is not a directory. Exiting...");
      System.exit(-1);
    }

    Path outputPath;
    if (optionOutputDir.isMatched()) {
      outputPath = Paths.get(optionOutputDir.value());
    } else {
      outputPath = Paths.get("./gen/");
      printMessage("No output dir is set. Assuming '" + outputPath.toAbsolutePath().toString() + "'.");
    }

    if (outputPath.toFile().exists() && !outputPath.toFile().isDirectory()) {
      printMessage("Output path '" + inputPath.toAbsolutePath().toString() + "' exists, but is not a directory. Exiting...");
    }

    printMessage("Running RelAST Preprocessor");

    Program program = parseProgram(inputPath);

    printMessage("Writing output files");

    for (GrammarFile grammarFile : program.getGrammarFileList()) {
      // TODO decide and document what the file name should be, the full path or a simple name?
      writeToFile(outputPath + grammarFile.getFileName() + "/Grammar.relast", grammarFile.generateAbstractGrammar());
    }

    writeToFile(outputPath + "/Grammar.relast", program.generateAbstractGrammar());
    return 0;
  }

  private void printMessage(String message) {
    System.out.println(message);
  }

  private void writeToFile(String filename, String str) throws CompilerException {
    try {
      PrintWriter writer = new PrintWriter(filename);
      writer.print(str);
      writer.close();
    } catch (Exception e) {
      throw new CompilerException("Could not write to file " + filename, e);
    }
  }

  private Program parseProgram(Path inputPath) throws CompilerException {
    Program program = new Program();

    try (DirectoryStream<Path> stream = Files.newDirectoryStream(inputPath, "*.relast")) {
      RelAstParser parser = new RelAstParser();
      stream.forEach(path -> {
        try (BufferedReader reader = Files.newBufferedReader(path)) {
          RelAstScanner scanner = new RelAstScanner(reader);
          GrammarFile inputGrammar = (GrammarFile) parser.parse(scanner);
          inputGrammar.dumpTree(System.out);
          program.addGrammarFile(inputGrammar);
          inputGrammar.treeResolveAll();
        } catch (IOException | Parser.Exception e) {
          printMessage("Could not parse grammar file " + path);
          e.printStackTrace();
        }
      });
    } catch (IOException e) {
      printMessage("Unable to iterate over input path '" + inputPath.toAbsolutePath().toString() + "'. Exiting...");
      e.printStackTrace();
      System.exit(-1);
    }


    return program;
  }
}

